from argparse import ArgumentParser, ArgumentTypeError, Action
import json
import logging
import os

from . import __version__
from .converter import convert_tiles, convert_hierarchy
from .cesium.config import cesium_settings_from_entwine_config
from enum import Enum


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def main():
  parser = ArgumentParser(prog='entium', description='Convert the entwine hierarchy to a cesium tileset')

  class FullPaths(Action):
    def __call__(self, parser, namespace, values, option_string=None):
      setattr(namespace, self.dest, os.path.abspath(os.path.expanduser(values)))

  def is_dir(dirname):
    if not os.path.isdir(dirname):
      msg = '{0} is not a directory'.format(dirname)
      raise ArgumentTypeError(msg)
    else:
      return dirname

  def is_json(filename):
    if filename is not None and not os.path.isfile(filename):
      raise ArgumentTypeError('{0} is not a file'.format(filename))
    if not os.path.splitext(filename)[1] == '.json':
      raise ArgumentTypeError('{0} is not a json file'.format(filename))

    return filename

  parser.add_argument('mode', choices=['tileset', 'tile', 'both'])
  parser.add_argument('entwine_dir', action=FullPaths, type=is_dir, help='input folder for entwine')
  parser.add_argument('output_dir', action=FullPaths, type=is_dir, help='output folder for the cesium tilests')
  parser.add_argument('-p', '--precision', nargs='?', type=float, default=0.01, help='precision in meters required to use quantized tiles')
  parser.add_argument('-c', '--config', action=FullPaths, nargs='?', type=is_json, help='filepath to config file to use advanced features')
  parser.add_argument('--validate', action='store_true', help='run post-process to validate point precision')
  parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))

  args = parser.parse_args()

  groups, batched = None, None
  if args.config is not None:
    with open(args.config, 'r') as config_file:
      config = json.load(config_file)

      groups, batched = cesium_settings_from_entwine_config(config)

  # TODO - Multithread
  if args.mode == 'both' or args.mode == 'tile':
    logger.info('Converting tiles...')
    convert_tiles(args.entwine_dir, args.output_dir, args.precision, args.validate, groups, batched)

  if args.mode == 'both' or args.mode == 'tileset':
    logger.info('Generating tileset hierarchy...')
    convert_hierarchy(args.entwine_dir, args.output_dir)

if __name__ == '__main__':
  main()


import glob
import json
import logging
import os

from .cesium.tiles import create_pointcloud, Mode, BatchComponentType, QUANTIZED_ECEF_CONSTANT
from .cesium.tileset import DirectTile, ReferenceTile
from enum import Enum, IntEnum
import numpy as np


logger = logging.getLogger(__name__)

def get_tileset_json(header, root_directory, global_meta):
  tileset = {}

  def _find_children(tile):
    if tile.depth + 1 not in tileset:
      return []
    x = tile.x * 2
    y = tile.y * 2
    z = tile.z * 2

    def is_within_cartesian(test_tile):
      return x <= test_tile.x < x + 2 and y <= test_tile.y < y + 2 and z <= test_tile.z < z + 2

    return filter(is_within_cartesian, tileset[tile.depth + 1])

  def _link_children(parents):
    for parent in parents:
      if isinstance(parent, ReferenceTile):
        continue
      parent.children = _find_children(parent)
      _link_children(parent.children)
    return parents
  
  # Get basic info on depth requirements
  base_depth = int(header.split('-')[0])
  step_size = 0 if 'hierarchyStep' not in global_meta else global_meta['hierarchyStep']

  tileset_path = os.path.join(root_directory, 'h', header)
  with open(tileset_path) as data_file:
    data = json.load(data_file)
    for tile_file in data.keys():
      tile_meta = map(int, tile_file.split('.')[0].split('-'))
      depth = tile_meta[0]
      if depth not in tileset:
        tileset[depth] = []
      is_reference = step_size != 0 and depth != base_depth and depth % step_size == 0
      tileset[depth].append(ReferenceTile(*tile_meta) if is_reference else DirectTile(*tile_meta))

  # Find all children at 
  root = _link_children(tileset[base_depth])[0]

  return {
    'asset': {
      'version': '0.0'
    },
    'geometricError': root.get_geometric_error(global_meta),
    'root': root.get_json(global_meta)
  }

def convert_hierarchy(input_path, output_path):
  if not os.path.isdir(input_path):
    raise 'Path provided is not a directory'
  
  logger.info('Reading meta...')
  with open(os.path.join(input_path, 'entwine.json'), 'r') as meta_file:
    meta = json.load(meta_file)

  headers_path = os.path.join(input_path, 'h')
  for header in os.listdir(headers_path):
    if not os.path.isfile(os.path.join(headers_path, header)):
      logger.warning('Skipping! %s' % header)
      continue

    header_id = int(header.split('-')[0])
    name = 'tileset.json' if header_id is 0 else 'tileset-' + header
    logging.info('Creating %s' % name)
    data = get_tileset_json(header, input_path, meta)
    with open(os.path.join(output_path, name), 'w') as outfile:
      logging.info('Writing %s'  % name)
      json.dump(data, outfile, indent=4)
      logging.info('Finished %s' % name)

def import_entwine_table(input_path, batch_header, groups, batched):
  entwine_header_dtype = np.dtype([ (x['name'], x['type'].value) for x in batch_header ])
  with open(input_path, 'rb') as raw_tile:
    content = np.fromfile(raw_tile, dtype=entwine_header_dtype)
    tile = create_pointcloud(content, mode=Mode.QUANTIZED, groups=groups, batch_columns=batched)
    if 'OriginId' in tile.batch_table:
      tile.batch_table.remove('OriginId') # Remove origin ID (artifact from cesium) when present
    return tile

# Cesium does not support > 16 bit integers, store bytes in alternate
class EntwineScemaType(Enum):
  INT8 = BatchComponentType.BYTE 
  INT16 = BatchComponentType.SHORT
  UINT8 = BatchComponentType.UNSIGNED_BYTE 
  UINT16 = BatchComponentType.UNSIGNED_SHORT
  FLOAT = BatchComponentType.FLOAT
  #UINT32 = BatchComponentType.FLOAT # Temporarily disabled for lack of webgl support
  #INT32 = BatchComponentType.FLOAT # Temporarily disabled for lack of webgl support
  #INT64 = BatchComponentType.DOUBLE # Temporarily disabled for lack of webgl support
  #UINT64 = BatchComponentType.DOUBLE Temporarily disabled for lack of webgl support 
  #DOUBLE = BatchComponentType.DOUBLE # Temporarily disabled for lack of webgl support

# Get header
def get_schema_type(name, type):
  raw_schema_type = str(type).strip().upper()
  if name in ['X', 'Y', 'Z'] and raw_schema_type == 'DOUBLE':
    return BatchComponentType.DOUBLE # Allow an exception for the double type for position data
  if name == 'OriginId':
    return BatchComponentType.UNSIGNED_INT
  if raw_schema_type not in EntwineScemaType.__members__:
    raise Exception('Unknown schema type: %s (%s)' % (raw_schema_type, name))
  return EntwineScemaType[raw_schema_type].value

def convert_tiles(input_path, export_path, precision=None, validate=False, groups=None, batched=None):
  total_points, total_tiles, high_precision_tiles = 0, 0, 0
  with open(os.path.join(input_path, 'entwine.json'), 'r') as meta_file:
    metadata = json.load(meta_file)
    header = [ { 'name': str(x['name']), 'type': get_schema_type(x['name'], x['type']) } for x in metadata['schema'] ]

  
  for bin_file in glob.iglob(os.path.join(input_path, '*.bin')):
    logging.info('Converting %s' % bin_file)
    tile = import_entwine_table(bin_file, header, groups, batched)
  
    points_column = tile.points
    if np.any((tile.bounds['max'] - tile.bounds['min']) / QUANTIZED_ECEF_CONSTANT > precision):
      tile.mode = Mode.FLOATING_QUANTIZED
      high_precision_tiles += 1

    cesium_file_name = '%s.pnts' % os.path.splitext(os.path.basename(bin_file))[0]
    cesium_file_path = os.path.join(export_path, cesium_file_name)
    tile.save(cesium_file_path)

    if validate:
      if points_column.mode is Mode.RTC_CENTER:
        converted_points = points_column.data() + points_column.rtc_point
      else:
        multiplier = points_column.quantized_scale 
        multiplier *= 1.0 / QUANTIZED_ECEF_CONSTANT if points_column.mode is Mode.QUANTIZED else 1.0
        converted_points = (points_column.data() * multiplier) + points_column.bounds['min']
      
      original_points = points_column.data
      distances = np.abs(converted_points - original_points)
      for idx in xrange(len(original_points)):
        distance = distances[idx]
        if np.any(distance > 1):
          original = original_points[idx]
          converted = converted_points[idx]
          logging.warning('\t- Outside Tolerance ' + str(original) + ' ~ ' + str(np.abs(original - converted)))


    total_points += tile.total_points
    total_tiles += 1

  logging.info('Completed Tiling')
  logging.info('\t- Tiles {:,}'.format(total_tiles))
  logging.info('\t- High Precision Tiles {:,}'.format(high_precision_tiles))
  logging.info('\t- Points {:,}'.format(total_points))



import json
from math import sqrt
import os


class Tile(object):

  def __init__(self, depth, x, y, z):
    self.x = x
    self.y = y
    self.z = z
    self.depth = depth

  def get_content_url(self, meta):
    raise NotImplementedError('Must override function!')

  def __repr__(self):
    return '[x %d, y %d, z %d, d %d]' % (self.x, self.y, self.z, self.depth)

  def _localize_bounds(self, meta, min_size=5000):
    scale_factor = pow(2, self.depth)
    bounds = meta['bounds']
    dimensions = [ abs(bounds[idx] - bounds[idx + 3]) / scale_factor for idx in xrange(0, 3) ]

    min_dimensions = [ max(min_size, x) for x in dimensions ]
    return {
      'x': bounds[0] + (dimensions[0] * self.x + (dimensions[0] / 2)),
      'y': bounds[1] + (dimensions[1] * self.y + (dimensions[1] / 2)),
      'z': bounds[2] + (dimensions[2] * self.z + (dimensions[2] / 2)),
      'width': min_dimensions[0],
      'depth': min_dimensions[1],
      'height': min_dimensions[2]
    }

  def get_geometric_error(self, meta):
    bounds = self._localize_bounds(meta)
    return sqrt(bounds['width']**2 + bounds['depth']**2 + bounds['height']**2) / 2

  def get_json(self, meta):
    bounds = self._localize_bounds(meta)
    return {
      'content': {
        'uri': self.get_content_url()
      },
      'refine': 'ADD',
      'geometricError': self.get_geometric_error(meta),
      'boundingVolume': {
        'box': [
          bounds['x'], bounds['y'], bounds['z'],     # Center
          bounds['width'], 0, 0, # X Transform
          0, bounds['depth'], 0, # Y Transform
          0, 0, bounds['height'] # Z Transform
        ]
      }
    }

class DirectTile(Tile):

  def __init__(self, depth, x, y, z, extension='pnts', children=None):
    super(DirectTile, self).__init__(depth, x, y, z)

    if children is None:
      children = []

    self.extension = extension
    self.children = children

  def get_json(self, meta):
    serialized = Tile.get_json(self, meta)
    if (len(self.children) > 0):
      serialized['children'] = [ x.get_json(meta) for x in self.children ]
    return serialized

  
  def get_content_url(self):
    return '%d-%d-%d-%d.%s' % (self.depth, self.x, self.y, self.z, self.extension)

class ReferenceTile(Tile):

  def __init__(self, depth, x=0, y=0, z=0):
    Tile.__init__(self, depth, x, y, z)
  
  def get_content_url(self):
    return 'tileset-%d-%d-%d-%d.json' % (self.depth, self.x, self.y, self.z)



def cesium_settings_from_entwine_config(config):
  def asciify(accum, x):
    if isinstance(x[1], list):
      remapped = [ y.encode('ascii', 'ignore') for y in x[1] ]
    else:
      remapped = x[1].encode('ascii', 'ignore')
    accum[x[0].encode('ascii', 'ignore')] = remapped
    return accum

  cesium_config = config['cesium']

  groups, batched = None, None
  if 'groups' in cesium_config:
    groups = reduce(asciify, cesium_config['groups'].iteritems(), {})

  if 'batched' in cesium_config:
    batched =  [ y.encode('ascii', 'ignore') for y in cesium_config['batched'] ]

  return groups, batched



import glob
import json
import os
import struct

from enum import IntEnum, Enum
import numpy as np
from numpy.lib.recfunctions import merge_arrays


def get_padding_bytes(total_bytes, required_multiple):
  if total_bytes % required_multiple != 0:
    return required_multiple - (total_bytes % required_multiple)
  else:
    return 0 

def binjsonify(func):
  def wrapper(*args, **kwargs):
    result = func(*args, **kwargs)
    json_dump = json.dumps(result, separators=(',', ':'))
    # Required 4  byte Padding for parsing of data (https://github.com/AnalyticalGraphicsInc/3d-tiles/blob/master/TileFormats/BatchTable/README.md#implementation-notes)
    # Doubled to 8 since 8 is the largest amount of bytes that can be stored
    json_dump += ' ' * get_padding_bytes(len(json_dump), 8) 
    return json_dump.encode('utf-8'); 
  return wrapper

class BatchComponentType(Enum):
  BYTE = np.int8 
  UNSIGNED_BYTE = np.uint8
  SHORT = np.int16
  UNSIGNED_SHORT = np.uint16
  INT = np.int32
  UNSIGNED_INT = np.uint32
  FLOAT = np.float32
  DOUBLE = np.float64

class BatchType(Enum):
  SCALAR = 1
  VEC2 = 2
  VEC3 = 3
  VEC4 = 4
  
class Mode(Enum):
  STANDARD = 0
  QUANTIZED = 1
  FLOATING_QUANTIZED = 2
  RTC_CENTER = 3

# Required by cesium to properly scale
# https://github.com/AnalyticalGraphicsInc/3d-tiles/tree/master/TileFormats/PointCloud#quantized-positions
QUANTIZED_ECEF_CONSTANT = float(pow(2, 16) - 1)

class AbstractColumn(object):
  def __init__(self, name, data):
    self.name = name
    self._data = data
    self.dtype = self._data.dtype.type if self.count() == 1 else self._data.dtype[0].type

    if self.count() > 1 and any([ data.dtype[name].type != self.dtype for name in data.dtype.names ]):
      info = ', '.join([ '%s(%s)' % (x, str(data.dtype[x])) for x in data.dtype.names ])
      raise ValueError('Datatypes are not matching %s' % info)

  def __eq__(self, other):
    if isinstance(other, BatchColumn):
        return self.name == other.name
    elif isinstance(other, str):
      return self.name == other
    return False

  def data(self):
    return self._data

  def count(self):
    return 1 if self._data.dtype.names is None else len(self._data.dtype.names)

  def names(self):
    if self.count() == 1:
      return [ self.name ]
    else:
      return self._data.dtype.names
  
  def get_itemsize(self):
    return self.data().itemsize

  def get_size(self):
    return self.data().nbytes

  def get_header(self, offset):
    raise NotImplementedError('get_header has not been implemented!')


class BatchColumn(AbstractColumn):
  def __init__(self, name, data, is_instanced=False):
    super(BatchColumn, self).__init__(name, data)
    self.is_instanced = is_instanced

  def get_component_type(self):
    return BatchComponentType(self.dtype)

  def get_batch_type(self):
    return BatchType(self.count())

  def get_header(self, offset):
    if self.is_instanced:
      if self.count() == 1:
        content = self._data.tolist()
      else:
        content = { name: self._data[name].tolist() for name in self._data.dtype.names }
    else:
      content = {
        'byteOffset': offset,
        'componentType': self.get_component_type().name,
        'type': self.get_batch_type().name
      }

    return { self.name: content }

class FeatureColumn(AbstractColumn):

  # Feature types and what is batch_item
  # https://github.com/AnalyticalGraphicsInc/3d-tiles/blob/master/TileFormats/PointCloud/README.md#point-semantics
  TYPES = {
    'position': 3,
    'rgb': 3,
    'rgba': 4,
    'rgb565': 1,
    'normal': 3,
    'normal_oct16p': 2,
    'batch_id': 1
  }

  def __init__(self, name, data, header_semantics=None):
    super(FeatureColumn, self).__init__(name, data)
    
    if header_semantics is None:
      header_semantics = {}

    if self.count() != FeatureColumn.TYPES[name]:
      raise Exception('Expected %d items in %s but recieved %d.' % (FeatureColumn.TYPES[name], name, self.count()))
    
    self.header_semantics = header_semantics

  def get_header(self, offset):
    header = {
      self.name.upper(): {
        'byteOffset': offset
      }
    }

    header.update(self.header_semantics)

    return header

class PositionColumn(FeatureColumn):
  def __init__(self, name, data, mode):
    super(PositionColumn, self).__init__(name, data)
    self.length = self._data.size # Save the size before transform
    self._data = self._data.view((self.dtype, 3))
    self.mode = mode

  def get_header(self, offset):
    if self.mode is Mode.RTC_CENTER or self.mode is Mode.STANDARD:
      header = {
        'POSITION': {
          'byteOffset': offset
        }
      }

      if self.mode is Mode.RTC_CENTER:
        header['RTC_CENTER'] = self.rtc_point.tolist()

    else:
      header = {
        'QUANTIZED_VOLUME_SCALE': self.quantized_scale.tolist(),
        'QUANTIZED_VOLUME_OFFSET': self.bounds['min'].tolist()
      }

      if self.mode is Mode.FLOATING_QUANTIZED:
        header['FLOATING_POSITION_QUANTIZED'] = {
          'byteOffset': offset
        }
      else:
        header['POSITION_QUANTIZED'] = {
          'byteOffset': offset
        }

    header['POINTS_LENGTH'] = self.length
    return header

  @property
  def bounds(self):
    return {
      'min': np.min(self._data, axis=0),
      'max': np.max(self._data, axis=0)
    }

  def data(self):
    if self.mode is Mode.RTC_CENTER:
      return self.rtc_points
    elif self.mode is Mode.QUANTIZED:
      return self.quantized_points
    elif self.mode is Mode.FLOATING_QUANTIZED:
      return self.normalized_points

  @property
  def rtc_point(self):
    return self.bounds['min'] + (self.bounds['max'] / 2.0)

  @property
  def rtc_points(self):
    return (self._data - self.rtc_point) \
      .astype(np.float32)

  @property
  def quantized_scale(self):
    return np.absolute(self.bounds['max'] - self.bounds['min']) 
      
  @property
  def quantized_points(self):
    multiplier = QUANTIZED_ECEF_CONSTANT / self.quantized_scale
    return ((self._data - self.bounds['min']) * multiplier) \
      .astype(np.uint16)

  @property
  def normalized_points(self):
    return np.nan_to_num((self._data - self.bounds['min']) / self.quantized_scale) \
      .astype(np.float32)

class Table(list):

  @binjsonify
  def get_header(self):
    table = {}
    offset = 0
    for item in self:
      offset += get_padding_bytes(offset, item.get_itemsize())
      table.update(item.get_header(offset))
      offset += item.get_size()

    return table

  def get_size(self):
    return reduce(lambda offset, item: offset + get_padding_bytes(offset, item.get_itemsize()) + item.get_size(), self, 0)

  def write(self, write_buffer, byte_offset=0):
    # Write Batch Table
    write_buffer.write(self.get_header())
    # Write each property sequentially
    for item in self:
      # Write padding to fix offset (Required that data starts on multiple of byte size to be parsed in JS)
      padding = get_padding_bytes(byte_offset, item.get_itemsize())
      write_buffer.write(struct.pack('x' * padding))
      byte_offset += padding

      item.data().tofile(write_buffer)
      byte_offset += item.get_size()


DEFAULT_GROUPS = { 'position': ['X', 'Y', 'Z'] }

def merge_dicts(x, y):
    z = x.copy() # start with x's keys and values
    z.update(y)  # modifies z with y's keys and values & returns None
    return z

def create_pointcloud(data, mode=None, groups=None, batch_columns=None):
  if mode is None:
    mode = Mode.STANDARD
  if groups is None:
    groups = {}
  groups = merge_dicts(DEFAULT_GROUPS, groups)
  if batch_columns is None:
    batch_columns = []

  columns = []
  def add(name, data):
    if name == 'position':
      columns.append(PositionColumn(name, data, mode))
    elif name.lower() in FeatureColumn.TYPES:
      columns.append(FeatureColumn(name.lower(), data))
    else:
      columns.append(BatchColumn(name, data))


  # Remap columns based off their groupings
  grouped = set()
  for name, selection in groups.iteritems():
    add(name, data[selection])
    grouped.update(selection if isinstance(selection, list) else [selection])
  for column in (set(data.dtype.names) - grouped):
    add(column, data[column])

  # Find all mapped values and replace it with their mapping
  if len(batch_columns) > 0:
    r_columns, r_names = [], []
    for column in batch_columns:
      if column not in columns:
        raise Exception('Column %s does not exist' % column)
      r_column = columns[columns.index(column)]
      r_columns.append(r_column)
      r_names.extend(r_column.names())

    merged_data = merge_arrays([ x.data() for x in r_columns ], flatten=True, usemask=False)
    merged_data.dtype.names = r_names
    batch_groups, batch_ids = np.unique(merged_data, axis=0, return_inverse=True)
    batch_ids = batch_ids.astype(np.uint16)

    idx_offset = 0
    for column in r_columns:
      column.is_instanced = True
      d_names = column.names()
      selector = d_names[0] if column.count() == 1 else d_names
      column._data = batch_groups[selector]

    columns.append(FeatureColumn('batch_id', batch_ids, { 'BATCH_LENGTH': len(batch_groups) }))

  return PointcloudTile(columns)

class PointcloudTile(object):

  def __init__(self, columns):
    self.total_points = -1
    self.position_column = None
    self.feature_table = Table()
    self.batch_table = Table()

    for column in columns:
      if isinstance(column, PositionColumn):
        self.position_column = column
      if isinstance(column, FeatureColumn):
        self.feature_table.append(column)
      elif isinstance(column, BatchColumn):
        self.batch_table.append(column)
      else:
        raise Exception('Unknown column type!')

    # Validate the minimum amount of data is present
    if self.position_column is None:
      raise Exception('Position is not present!')

    self.total_points = self.position_column.length

  @property
  def points(self):
    return self.position_column
  
  @property
  def bounds(self):
    return self.points.bounds

  @property
  def mode(self):
    self.points.mode

  @mode.setter
  def mode(self, mode):
    self.points.mode = mode

  def save(self, output_path):
    with open(output_path, 'wb') as cesium_tile:
      header_struct = struct.Struct('4sIIIIII')

      feature_header = self.feature_table.get_header()
      feature_size = self.feature_table.get_size()
      padding = get_padding_bytes(header_struct.size + feature_size + len(feature_header), 8)

      # Skip writing if size is 0
      if len(self.batch_table) > 0:
        batch_header_length = len(self.batch_table.get_header())
        batch_size = self.batch_table.get_size()
      else:
        batch_header_length = 0
        batch_size = 0

      # Write Header
      cesium_tile.write(header_struct.pack(
        'pnts',                  # magic key (DO NOT CHANGE)
        1,                       # Version, It has to be one according ot docs
        0,                       # Byte length (currently unread)
        len(feature_header),     # byte space of json info
        feature_size + padding,  # byte space of feature data, we include padding as it will be excluded later
        batch_header_length,     # byte space of json info
        batch_size              # byte space of batch data
      ))

      self.feature_table.write(cesium_tile) # Write Feature Table
      cesium_tile.write(struct.pack('x' * padding)) # Write Padding
      if len(self.batch_table) > 0:
        self.batch_table.write(cesium_tile) # Write Batch table




from codecs import open
from os.path import abspath, dirname, join
from subprocess import call

from setuptools import Command, find_packages, setup

from entium import __version__


project_dir = abspath(dirname(__file__))
with open(join(project_dir, 'README.md'), encoding='utf-8') as file:
  long_description = file.read()

with open(join(project_dir, 'requirements.txt')) as file:
  required = file.read().splitlines()

setup(
  name='entium',
  version= __version__,
  keywords='cli, entwine, cesium, converter',
  description='A command line tool to read entwine\'s convert into Cesium 3DTiles',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url='https://github.com/commaai/entium',
  author='Brandon Barker',
  author_email='brandon@comma.ai',
  license='MIT LICENSE',
  packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
  install_requires=required,
  classifiers=[
    'Intended Audience :: Developers',
    'Topic :: Utilities',
    'License :: Public Domain',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 2.7'
  ],
  entry_points={
    'console_scripts': [
      'entium=entium.__main__:main',
    ]
  }
)
